#include <types.h>
#include <memory.h>
#include <device.h>
#include <interrupt.h>
#include <time.h>
#include <lib/string.h>

#include <dev/waveapi.h>
#include <dev/wavedrv.h>

#include <mach/hal_interrupt.h>
#include <mach/AT91SAM9GX5.h>
#include <mach/AT91SAM9GX5bd.h>

#include <mach/gpio.h>

#include "ac97.h"
#include "wave.h"

#include <hal_panic.h>

#define DBG_LVL_AC97    0

PUBLIC DEVICECONFIGWAVE        WAV_DeviceConfig ={"WAV", 0, -1, NULL, NULL, SYSINTR_WAV };

//==============================================================================
// hardware dependent definition
//==============================================================================

#define AC97FS	AT91C_PIN_PD(8)
#define AC97CK	AT91C_PIN_PD(9)
#define AC97TX	AT91C_PIN_PD(7)
#define AC97RX	AT91C_PIN_PD(6)

STATIC  AT91PS_AC97C    g_pAC97C = (AT91PS_AC97C)AT91C_BASE_AC97C;
STATIC  AT91PS_PMC      g_pPMC = (AT91PS_PMC)AT91C_BASE_PMC;
#define READ_TIMEOUT    10
#define WRITE_TIMEOUT   10

//STATIC UINT32  g_dwCodecType=GENERIC_AC97;
STATIC UINT16           LastSampleRateIn=KHZ48_000;
STATIC UINT16           LastSampleRateOut=KHZ48_000;

STATIC UINT16           g_VolMatrix[0x64];
STATIC UINT32           v_nVolume=0xeeeeeeee; //default to volume
STATIC BOOL             bPowerDown[2] = {TRUE, TRUE };


#define AUDIO_BUFFER_SIZE           4096


STATIC PBYTE            DMAPage[NUM_DMA_AUDIO_BUFFERS];





//==============================================================================
// from here, add the hardware specific routines
//==============================================================================




STATIC BOOL
ConfigureAudioPIOs( VOID )
{    
	const struct pio_desc hw_pio[] = 
    {
        {"AC97FS",	AC97FS  , 0, PIO_PULLUP, PIO_PERIPH_A},
        {"AC97CK",	AC97CK  , 0, PIO_PULLUP, PIO_PERIPH_A},
        {"AC97TX",	AC97TX  , 0, PIO_PULLUP, PIO_PERIPH_A},
        {"AC97RX",	AC97RX	, 0, PIO_PULLUP, PIO_PERIPH_A},
    };
    pio_setup(hw_pio, sizeof(hw_pio)/sizeof(struct pio_desc));

    return TRUE;
}










//-----------------------------------------------------------------------------
//! \fn            BOOL ReadAC97( UINT8 Offset, UINT16 *pData)
//!
//! \brief        read data on the AC97
//!
//! \param        Offset    Address where to read
//! \param        pData    Address where to write
//!
//! \return        TRUE when all is good
//!    \return        FALSE when all is bad
//-----------------------------------------------------------------------------
STATIC BOOL
ReadAC97( UINT8 Offset, PUINT16 pData )
{
    UINT64 eTime;
    UINT64 sTime;

    UINT32 dwDummy = g_pAC97C->AC97C_COSR; //Clear th RXRDY bit if set

    if ((pData == NULL) || (Offset > 0x7f))
    {
        return FALSE;
    }

    MsCountGet( &eTime );
    eTime += READ_TIMEOUT;
    do
    {
        if (g_pAC97C->AC97C_COSR & AT91C_AC97C_TXRDY)
        {
            g_pAC97C->AC97C_COTHR = AT91C_AC97C_READ + (Offset << 16);
            break;
        }
        MsCountGet( &sTime );
    }while(eTime > sTime );


    MsCountGet( &eTime );
    eTime += READ_TIMEOUT;
    do
    {
        if (g_pAC97C->AC97C_COSR & AT91C_AC97C_RXRDY)
        {
            *pData = g_pAC97C->AC97C_CORHR;
            return TRUE;
        }
        MsCountGet( &sTime );
    }while(eTime > sTime );

    return(FALSE);
}


//-----------------------------------------------------------------------------
//! \fn            BOOL WriteAC97(UINT8 Offset, UINT16 wData)
//!
//! \brief        write data on the AC97
//!
//! \param        Offset    Address where to write
//! \param        wData    Data to write
//!
//! \return        TRUE when all is good
//!    \return        FALSE when all is bad
//-----------------------------------------------------------------------------
STATIC BOOL
WriteAC97(UINT8 Offset, UINT16 wData)
{
    UINT64 eTime;
    UINT64 sTime;
    
    if (Offset > 0x7f)
    {
        return FALSE;
    }
    
    MsCountGet( &eTime );
    eTime += WRITE_TIMEOUT;
    do
    {
        if (g_pAC97C->AC97C_COSR & AT91C_AC97C_TXRDY)
        {
            g_pAC97C->AC97C_COTHR = (Offset << 16) | wData;
            return TRUE;
        }
        MsCountGet( &sTime );
    }while(eTime > sTime );
    
    return(FALSE);
}



//------------------------------------------------------------------------------------------------------------
//! \brief Perform AC97 write and copy the write to a shadow register.  This helps performance by reducing the number of reads over the "slow" ac97 link
//!
//! \returns SUCCESS if there was a successful write to the AC97 codec
//!
//-------------------------------------------------------------------------------------------------------------
#define MAX_NB_TRY    3
STATIC UINT16 
ShadowWriteAC97( UINT8 Offset, UINT16 Data )
{
    //there are some values which should not be shadowed since the codec is volatile
    //TODO: return failure or warning on attempts to write values that shouldn't be shadowed
    
    BOOL    bRet;
    INT32   i = 0;
    
    do
    {
        if( (bRet = WriteAC97( Offset, Data ) ) == TRUE )
            break;

        i++;
    } while (i < MAX_NB_TRY);
    
    return bRet;
}


//------------------------------------------------------------------------------------------------------------
//! \brief \return a value from a shadowed ac97 codec register.
//-------------------------------------------------------------------------------------------------------------
UINT16 ShadowReadAC97(BYTE Offset, PUINT16 pData)
{
    BOOL    bRet;
    INT32   i = 0;
    
    do 
    {
        if( (bRet = ReadAC97(Offset,pData)) == TRUE )
            break;

        i++;
    } while (i < MAX_NB_TRY);
    
    return bRet;
}



//------------------------------------------------------------------------------------------------------------
//! \brief Quick Mute the output with out effecting the stored volume value
//!
//! \return TRUE indicates success. FALSE indicates failure
//!
//-------------------------------------------------------------------------------------------------------------
BOOL
WaveOutMute( BOOL bMute )
{
    
    UINT16  Ac97Data=0;
    
    ShadowReadAC97(HEADPHONE_VOLUME,&Ac97Data);  //read the current hardware volume
    
    if( bMute )
        Ac97Data = Ac97Data |   MUTE_MASK;      //set the mute bit
    else
        Ac97Data = Ac97Data & ~ MUTE_MASK;      //clear the mute bit
    
    
    return( ShadowWriteAC97(HEADPHONE_VOLUME, Ac97Data) );
}

//------------------------------------------------------------------------------------------------------------
//! \brief Quick Mute the input with out effecting the stored volume value
//!
//! \return TRUE indicates success. FALSE indicates failure
//!
//-------------------------------------------------------------------------------------------------------------
BOOL
WaveInMute( BOOL bMute )
{
    
    UINT16  Ac97Data=0;
    
    ShadowReadAC97(RECORD_GAIN,&Ac97Data);      //read the current hardware volume
    
    if( bMute )
        Ac97Data = Ac97Data |   MUTE_MASK;      //set the mute bit
    else
        Ac97Data = Ac97Data &  ~ MUTE_MASK;     //clear the mute bit
    
    
    return( ShadowWriteAC97(RECORD_GAIN, Ac97Data) );
}


//------------------------------------------------------------------------------------------------------------
//! \brief Convert the Windows Volume to an AC 97 volume and set it
//!
//! \return TRUE indicates success. FALSE indicates failure
//!
//-------------------------------------------------------------------------------------------------------------

STATIC BOOL
WaveOutSetVolume( UINT32 uVolume )
{
    UINT16  ucAC97VolL;
    UINT16  ucAC97VolR;
    UINT16  ucAC97VolBoth;
    UINT16  ucAC97VolCur;
    
    // Windows Volume is 16 bits of Left and 16 bits of right (0xRRRRLLLL)
    
    ucAC97VolR=(UINT16) (uVolume & 0x0000FFFF) >> 10; // six bits of volume
    ucAC97VolR=g_VolMatrix[ucAC97VolR];
    
    ucAC97VolL=(UINT16) ((uVolume & 0xFFFF0000)>>16) >> 10;  //six bits of volume
    ucAC97VolL=g_VolMatrix[ucAC97VolL];
    
    ucAC97VolBoth=(ucAC97VolL<<8) | ucAC97VolR;
    
    ShadowReadAC97( HEADPHONE_VOLUME, &ucAC97VolCur );
    
    ucAC97VolCur &= 0xC0C0;  //only keep the upper 2 bits of each byte
    ucAC97VolBoth &= ~0xC0C0;  //only keep lower 6 bits of each byte
    ucAC97VolBoth |= ucAC97VolCur;  //add current volume bites
    
    
    return (ShadowWriteAC97(HEADPHONE_VOLUME, ucAC97VolBoth));
    //return TRUE;
    
}
//------------------------------------------------------------------------------------------------------------
//! \brief Convert the Windows Volume to an AC 97 volume and set it
//!
//! \return TRUE indicates success. FALSE indicates failure
//!
//-------------------------------------------------------------------------------------------------------------

STATIC BOOL
WaveInSetVolume( UINT32 uVolume )
{
    UINT16 ucAC97VolL;
    UINT16 ucAC97VolR;
    UINT16 ucAC97VolBoth;
    UINT16 ucAC97VolCur;
    
    // Windows Volume is 16 bits of Left and 16 bits of right (0xRRRRLLLL)
    
    ucAC97VolR=(UINT16) (uVolume & 0x0000FFFF) >> 10; // six bits of volume
    ucAC97VolR=g_VolMatrix[ucAC97VolR];
    
    ucAC97VolL=(UINT16) ((uVolume & 0xFFFF0000)>>16) >> 10;  //six bits of volume
    ucAC97VolL=g_VolMatrix[ucAC97VolL];
    
    ucAC97VolBoth=(ucAC97VolL<<8) | ucAC97VolR;
    
    ShadowReadAC97( HEADPHONE_VOLUME, &ucAC97VolCur );
    
    ucAC97VolCur &= 0xC0C0;  //only keep the upper 2 bits of each byte
    ucAC97VolBoth &= ~0xC0C0;  //only keep lower 6 bits of each byte
    ucAC97VolBoth |= ucAC97VolCur;  //add current volume bites
    
    
    return (ShadowWriteAC97(HEADPHONE_VOLUME, ucAC97VolBoth));
    //return TRUE;
    
}



//------------------------------------------------------------------------------------------------------------
//! \brief Write the sample rate value to the ac97 codec
//-------------------------------------------------------------------------------------------------------------
STATIC BOOL
AC97SetSampleRate( UINT16 SampleRate, WAPI_INOUT apidir )
{
    BOOL    bRet = FALSE;
	
    switch(SampleRate)
    {
	case 8000:
	case 11025:
	case 16000:
	case 22050:
	case 32000:
	case 44100:
	case 48000:
		bRet=TRUE;
		break;

	default:
		bRet=FALSE;
		return bRet;
		
		
    }
	
    //this write assumes that the
    if (apidir == WAPI_IN)
    {
        bRet=ShadowWriteAC97((BYTE)AUDIO_ADC_RATE,(UINT16)(SampleRate & 0xFFFF)); //set the input sample rate
        LastSampleRateIn=SampleRate;
    }
    else
    {
        bRet=ShadowWriteAC97((BYTE)AUDIO_DAC_RATE,(UINT16)(SampleRate & 0xFFFF)); //set the output sample rate
        LastSampleRateOut=SampleRate;
    }
	
    return bRet;
}


//------------------------------------------------------------------------------------------------------------
//! \brief power up or power down a dac, adc, or other devices
//-------------------------------------------------------------------------------------------------------------
STATIC VOID
AudioPowerDownUnit( UINT16 Unit, BOOL bShutDown )
{
    register UINT16 usiMask,usiMaskedValue;
    UINT16          PowerDownCtrlStat=0;
    
    if( !ShadowReadAC97( POWERDOWN_CTRL_STAT, &PowerDownCtrlStat)  )  // Enable Variable Rate Audio
    {
DEBUGMSG(DBG_LVL_AC97, "-- read of PwrDnCtrl failed \n");
    }    

    usiMask = Unit;
    if (bShutDown)
    {
        usiMaskedValue = Unit;
    }
    else
    {
        usiMaskedValue = 0;
    }

    if ((PowerDownCtrlStat & usiMask) != usiMaskedValue)
    {
        PowerDownCtrlStat = (PowerDownCtrlStat & ~usiMask) | usiMaskedValue;
        
        if( !ShadowWriteAC97( POWERDOWN_CTRL_STAT, PowerDownCtrlStat)  )  // Enable Variable Rate Audio
        {
            DEBUGMSG(DBG_LVL_AC97,"-- write of PwrDnCtrl failed \n");
        }
    }
    
}


//------------------------------------------------------------------------------------------------------------
//! \brief Routine for powering on Audio hardware.  Reinit the codec
//!
//! \note This may be called from within power handler, so should not make any system calls.
//-------------------------------------------------------------------------------------------------------------
STATIC BOOL
AudioPowerOn( VOID )
{        
    STATIC UINT32 VendorId=0;
    UINT16 VidTemp=0;
    UINT16 toto;
    
    
    //Force VRA (variable rate audio) to be on

    // Enable Variable Rate Audio
    if( !ShadowWriteAC97( EXTENDED_AUDIO_CTRL, VRA_ENABLED_MASK)  )
    {     
        DEBUGMSG(DBG_LVL_AC97, "-- Bad VRA Value\n");
    }

    ShadowReadAC97(EXTENDED_AUDIO_CTRL, &toto);

    // default is AUDIO_MIC_INPUT_MONO
    if( !ShadowWriteAC97( RECORD_SELECT, 0x0 )  )                       
    {     
        DEBUGMSG(DBG_LVL_AC97, "-- Bad RECORD_SELECT\n"); 
    }

    //Set the record gain value
    //set to a well working value
    if( !ShadowWriteAC97( RECORD_GAIN, 0x0f0f )  )                      
    {     
        DEBUGMSG(DBG_LVL_AC97,"-- error setting RECORD_GAIN \n");    
    }
    
    //Mute every input of the mixer line 

    //Also add the 20db Microphone boost
    if( !ShadowWriteAC97( MIC_VOLUME, 0x8000 | (1<<6))  )               
    {     
        DEBUGMSG(DBG_LVL_AC97,"-- error setting PC_BEEP_VOLUME\n");    
    }
    
    //set to a well working value
    if( !ShadowWriteAC97( PC_BEEP_VOLUME, 0x8000 )  )                   
    {     
        DEBUGMSG(DBG_LVL_AC97,"-- error setting PC_BEEP_VOLUME\n");    
    }

    //set to a well working value    
    if( !ShadowWriteAC97( PHONE_VOLUME, 0x8000 )  )         
    {     
        DEBUGMSG(DBG_LVL_AC97,"-- error setting PHONE_VOLUME\n");    
    }

    //set to a well working value    
    if( !ShadowWriteAC97( LINE_IN_VOLUME, 0x8000 )  )
    {     
        DEBUGMSG(DBG_LVL_AC97,"-- error setting LINE_IN_VOLUME\n");    
    }


    //set to a well working value
    if( !ShadowWriteAC97( CD_VOLUME, 0x8000 )  )  
    {     
        DEBUGMSG(DBG_LVL_AC97,"-- error setting CD_VOLUME\n");    
    }

    //set to a well working value
    if( !ShadowWriteAC97( VIDEO_VOLUME, 0x8000 )  )  
    {     
        DEBUGMSG(DBG_LVL_AC97,"-- error setting VIDEO_VOLUME\n");    
    }

    //set to a well working value
    if( !ShadowWriteAC97( AUX_VOLUME, 0x8000 )  )  
    {     
        DEBUGMSG(DBG_LVL_AC97,"-- error setting AUX_VOLUME\n");    
    }

    //compliant codecs such as LM4549 require the PCM volume
    ShadowReadAC97(VENDOR_ID1, &VidTemp);           //ucb1400 doesn't use an ascii (that I can tell)
    VendorId= (VidTemp <<16);                       //ffffffffssssssss // f=first ascii s = second ascii
    ShadowReadAC97(VENDOR_ID2, &VidTemp);
    VendorId |=VidTemp;                             //ttttttttrrrrrrrr //t = third ascii, r=rev #
    VendorId &= 0xfffffff0;                         //trim of version number

    //vendor specific
    switch( VendorId )
    {            
    default:        
//        g_dwCodecType = GENERIC_AC97;

        //Enable 3D Sound ....
        if( !ShadowWriteAC97( GENERAL_PURPOSE, (1<<9) | (1<<13))  )
        {
//DEBUGMSG(DBG_LVL_AC97,"-- error setting GENERAL_PURPOSE\n");    
        }
        // Set the PCM out volume (which is not the master volume)
        if( !ShadowWriteAC97( PCM_OUT_VOL, 0x0404)  )
        {
//DEBUGMSG(DBG_LVL_AC97,"-- error setting PCM_OUT_VOL\n");    
        }
        break;                
    }
    
    AC97SetSampleRate( LastSampleRateOut, WAPI_OUT );
    AC97SetSampleRate( LastSampleRateIn, WAPI_IN );
    
    return (TRUE);
}

//------------------------------------------------------------------------------------------------------------
//! \brief Routine for powering down Audio hardware.
//-------------------------------------------------------------------------------------------------------------
STATIC VOID
AudioPowerOff( VOID )
{
  
    WavHal_WaveOutStop();
    WavHal_WaveInStop(NULL);
    

    WaveOutMute(TRUE);
    WaveInMute(TRUE);      
    
bPowerDown[WAPI_OUT] = TRUE;
bPowerDown[WAPI_IN] = TRUE;
    AudioPowerDownUnit(PWR_PR1_DAC, TRUE);
    AudioPowerDownUnit(PWR_PR0_ADC, TRUE);
    
}


//------------------------------------------------------------------------------------------------------------
//! \brief stops PDC (DMA) transfer
//-------------------------------------------------------------------------------------------------------------
STATIC BOOL  
StopDmac(WAPI_INOUT wapidir)
{
    switch(wapidir)
    {
	case WAPI_IN:
		g_pAC97C->AC97C_PTCR = AT91C_PDC_RXTDIS;
		break;
		
	case WAPI_OUT:	
		g_pAC97C->AC97C_PTCR = AT91C_PDC_TXTDIS;
		break;
		
	default:
		return FALSE;
	}
	
    return(TRUE);
}



STATIC BOOL
AC97_Initialize( VOID )
{
    int     i;
    BOOL    bRet = FALSE;


    
    //
    // Configure the PIOs
    //
    if (!ConfigureAudioPIOs())
    {
        return FALSE;
    }



    // Map both DMA pages into the local address space.
#if 0
    DMAPage[0]     = (PBYTE)DMA_NC_AUDIO;//0x3000d000;
#else
    //DMAPage[0]     = (PBYTE)DMA_AUDIO;
    DMAPage[0]     = (PBYTE)MemPageAlloc(4);
    DMAPage[0]    += 0x70000000;
    
#endif
    DMAPage[1]     = DMAPage[0] + AUDIO_BUFFER_SIZE;
    DMAPage[2]     = DMAPage[1] + AUDIO_BUFFER_SIZE;
    DMAPage[3]     = DMAPage[2] + AUDIO_BUFFER_SIZE;

    DEBUGMSG(DBG_LVL_AC97,"+WHAL_Init %X\n", DMAPage[0]);
    DEBUGMSG(DBG_LVL_AC97,"+WHAL_Init %X\n", DMAPage[1]);
    DEBUGMSG(DBG_LVL_AC97,"+WHAL_Init %X\n", DMAPage[2]);
    DEBUGMSG(DBG_LVL_AC97,"+WHAL_Init %X\n", DMAPage[3]);
    



    // Compress range of 3f - 0 to 1f - 0 this makes volume adjustment seem intuitive.
    //
    for (i=0; i < sizeof(g_VolMatrix) / sizeof(g_VolMatrix[0]); i+=2)
    {
        g_VolMatrix[i]   = (63 - i) >> 1;
        g_VolMatrix[i+1] = (63 - i) >> 1;
    }
    // Customize lower volume settings to quickly rolloff to neg infinity.
    g_VolMatrix[0] = 0x3f;
    g_VolMatrix[1] = 0x3A;
    g_VolMatrix[2] = 0x35;
    g_VolMatrix[3] = 0x2F;
    g_VolMatrix[4] = 0x2a;
    g_VolMatrix[5] = 0x25;
    g_VolMatrix[6] = 0x1f;
    
    
    
    //Turn on the controller clock;
    g_pPMC->PMC_PCER = (1<<AT91C_ID_AC97C);
    
    //Reset the audio codec 
    //ResetAC97Controller(CODEC_COLD_RESET);
    
    //Set up the AC97C registers
    g_pAC97C->AC97C_MR = AT91C_AC97C_ENA | AT91C_AC97C_VRA;
    
    //Selection of the input slot. We only manage Slot 3 (Mono Input only)
    g_pAC97C->AC97C_ICA = AT91C_AC97C_CHID3_CA;
    
    //Selection of the output slot. We only manage Slot 3 & 4 (PCM L and R Front)
    g_pAC97C->AC97C_OCA = AT91C_AC97C_CHID3_CA | AT91C_AC97C_CHID4_CA;
    
    
    
    
    // Initialisation of channel A (used for both input and output)
    // Configuration is : 
    // * PDC (dma) enabled
    // * Channel enabled
    // * Little endian 
    // * data width : 16 bits
    // * interrupts enabled for OVRUN and UNRUN      

    g_pAC97C->AC97C_CAMR = AT91C_AC97C_PDCEN | AT91C_AC97C_CEN | AT91C_AC97C_SIZE_16_BITS |  AT91C_AC97C_UNRUN;

    //Enable the interrupt for Channel A only
    g_pAC97C->AC97C_IDR = 0xFFFFFFFF;
    g_pAC97C->AC97C_IER = AT91C_AC97C_CAEVT;
    
    // Power-on the audio controller.
    if (AudioPowerOn())
    {
        bRet = TRUE;
    }

    //Set the default volume 
    WaveOutSetVolume(v_nVolume);

    
    return (bRet);
}






STATIC VOID
AC97_Deinitialize( VOID )
{
    // Disable DMA input.
    StopDmac(WAPI_IN);
    
    // Disable DMA output.
    StopDmac(WAPI_OUT);

    // Power-off the audio controller.
    AudioPowerOff();

    //Turn off the AC97C clock    
    g_pPMC->PMC_PCDR = (1<<AT91C_ID_AC97C);
    

    return;
}




//==============================================================================
//==============================================================================
//==============================================================================
//==============================================================================
//==============================================================================
//==============================================================================
//==============================================================================
//==============================================================================
// from here, HARDWARE INDEPENDENT CODE
//==============================================================================
//==============================================================================
//==============================================================================
//==============================================================================
//==============================================================================
//==============================================================================
//==============================================================================
//==============================================================================
//==============================================================================



#define NUM_API_DIRS          2


typedef struct {
    UINT16    wSrcBitsPerSample;
    UINT16    wDstBitsPerSample;
    UINT16    wSrcNbChannels;
    UINT16    wDstNbChannels;
    BOOL        bInUse;        // is apidir in use? if so, fail Open calls
    UINT32      dwDMASize; // Size in bytes
} T_CHANNEL_DESC;


STATIC T_CHANNEL_DESC   gChannelDesc[NUM_API_DIRS];
STATIC BOOL			    gbMoreData[NUM_API_DIRS];
STATIC UINT32           gnXmitComplete;             // number of byte already played by the codec

STATIC UINT32           gCaptureLatency;            // capture buffers will hold this much data (in milliseconds)

#define DEFAULT_CAPTURE_LATENCY         20



//==============================================================================
// buffer handle helper functions
//==============================================================================


typedef struct 
{
    PVOID   pData;
    UINT32  dwPhysAddr;
    UINT32  dwSize;
    UINT32  ID;
} T_BUFFER;

typedef struct 
{
    UINT32 dwListMaxSize;
    UINT32 dwNbBufferInList;
    UINT32 dwFirstBufferIndex;
    UINT32 dwNextBufferIndex;
    T_BUFFER *pBufferArray[MAX_BUFFER_PAGE];
} T_BUFFER_FIFO;


T_BUFFER_FIFO   FreeOutBufferFifo;          //! Fifo used to keep track of free emission buffer
T_BUFFER_FIFO   BufferToPlayFifo;           //! Fifo used to keep track of buffers ready to be played
T_BUFFER_FIFO   BufferBeingPlayedFifo;      //! Fifo used to keep track of buffers being played (already in the DMA list)
T_BUFFER        XmitBuffers[NB_XMIT_PAGE];

T_BUFFER_FIFO   FreeInBufferFifo;           //! Fifo used to keep track of buffer available for recording
T_BUFFER_FIFO   BufferRecordedFifo;         //! Fifo used to keep track of recorded buffers (after DMA)
T_BUFFER_FIFO   BufferBeingRecordedFifo;    //! Fifo used to keep track of buffers being recorded (already in the DMA list)
T_BUFFER        RcvBuffers[NB_RECV_PAGE];


STATIC VOID (*gpfnBufferFill) (PINT16 pDstBuffer, PVOID pSrcBuffer, UINT32 dwSamples);
STATIC VOID (*gpfnBufferGet)  (PVOID pDstBuffer, PINT16 pSrcBuffer, UINT32 dwSamples);

#define min(a,b)                    (a>b?b:a)

//-----------------------------------------------------------------------------
//! \brief        This function creates a new Fifo
//! \param        dwNbBuffer the number of buffer that the fifo can handle
//! \param        pFifo       a pointer to the fifo structure that will be initialized
//-----------------------------------------------------------------------------
STATIC BOOL
BufferFifoInit( T_BUFFER_FIFO* pFifo )
{
    if (pFifo == NULL)
    {
        return FALSE;
    }
    
    pFifo->dwListMaxSize = NB_XMIT_BUFF_FIFO;

    pFifo->dwNbBufferInList = 0;
    pFifo->dwFirstBufferIndex = 0;
    pFifo->dwNextBufferIndex = 0;
    
    return TRUE;
}


//-----------------------------------------------------------------------------
//! \brief        This function resets a fifo
//! \param        pFifo       a pointer to the fifo structure that will be reset
//!
//! The fifo is emptied. There's no need to create the fifo again. It's still ready to use
//-----------------------------------------------------------------------------
STATIC BOOL
BufferFifoReset(T_BUFFER_FIFO* pFifo)
{
    if (pFifo == NULL)
    {
        return FALSE;
    }
    
    pFifo->dwNbBufferInList = 0;
    pFifo->dwFirstBufferIndex = 0;
    pFifo->dwNextBufferIndex = 0;
    
    return TRUE;
}


//-----------------------------------------------------------------------------
//! \brief        This function deletes a fifo
//! \param        pFifo       a pointer to the fifo structure that will be deleted
//!
//! The fifo is emptied. All allocations are released.
//-----------------------------------------------------------------------------
STATIC BOOL 
BufferFifoDelete( T_BUFFER_FIFO* pFifo )
{
    if (pFifo == NULL)
    {
        return FALSE;
    }
    
    pFifo->dwNbBufferInList = 0;
    pFifo->dwFirstBufferIndex = 0;
    pFifo->dwNextBufferIndex = 0;    
    pFifo->dwListMaxSize = 0;

    return TRUE;
}


//-----------------------------------------------------------------------------
//! \brief        This function gets the first buffer of the list and remove it from the fifo
//! \param        pFifo       a pointer to the fifo structure that will be deleted
//!
//! \return  A pointer to the first buffer of the fifo if it's not empty, else NULL
//-----------------------------------------------------------------------------
T_BUFFER* 
BufferFifoPop(T_BUFFER_FIFO* pFifo)
{
    T_BUFFER* pResult = NULL;
    
    if (pFifo == NULL)
    {
        return NULL;
    }

    if (pFifo->dwNbBufferInList != 0)
    {
        pResult = pFifo->pBufferArray[pFifo->dwFirstBufferIndex];        
        pFifo->dwFirstBufferIndex = INC_MODULO(pFifo->dwFirstBufferIndex,pFifo->dwListMaxSize);        
        pFifo->dwNbBufferInList--;
    }

    return pResult;
}

//-----------------------------------------------------------------------------
//! \brief        This function gets the first buffer of the list and remove it from the fifo
//! \param        pFifo       a pointer to the fifo structure that will be deleted
//! \param        pBuffer        a pointer to the buffer that must be stored in the fifo
//!
//! \return  TRUE indicates success, FALSE indicates failure
//-----------------------------------------------------------------------------
STATIC BOOL
BufferFifoPush( T_BUFFER_FIFO* pFifo,T_BUFFER* pBuffer )
{
    BOOL bResult = FALSE;
    if (pFifo == NULL)
    {
        return FALSE;
    }

    if (pFifo->dwNbBufferInList < pFifo->dwListMaxSize)
    {
        pFifo->pBufferArray[pFifo->dwNextBufferIndex] = pBuffer;
        pFifo->dwNextBufferIndex = INC_MODULO(pFifo->dwNextBufferIndex,pFifo->dwListMaxSize);        
        pFifo->dwNbBufferInList++;
        bResult = TRUE;
    }

    return bResult;
}



STATIC VOID
BufferFillM08toS16 (PINT16 pDstBuffer, PBYTE pSrcBuffer, UINT32 dwSamples)
{
//DEBUGMSG(DBG_LVL_AC97,"BufferFillM08toS16\n");
    while (dwSamples-- > 0) 
    {
        PPCM_SAMPLE pSample = (PPCM_SAMPLE) pSrcBuffer;
        INT16 sample16 = BITS_8_TO_16(pSample->m8.sample);
        pDstBuffer[0] = sample16;
        pDstBuffer[1] = sample16;
        pDstBuffer += 2;
        pSrcBuffer += 1;
    }
}

STATIC VOID
BufferFillS08toS16 (PINT16 pDstBuffer, PBYTE pSrcBuffer, UINT32 dwSamples)
{
//DEBUGMSG(DBG_LVL_AC97,"BufferFillS08toS16\n");
    while (dwSamples-- > 0) 
    {
        PPCM_SAMPLE pSample = (PPCM_SAMPLE) pSrcBuffer;
        pDstBuffer[0] = BITS_8_TO_16(pSample->s8.sample_left);
        pDstBuffer[1] = BITS_8_TO_16(pSample->s8.sample_right);
        pDstBuffer += 2;
        pSrcBuffer += 2;
    }
}

STATIC VOID
BufferFillM16toS16 (PINT16 pDstBuffer, PINT16 pSrcBuffer, UINT32 dwSamples)
{
//DEBUGMSG(DBG_LVL_AC97,"BufferFillM16toS16\n");
    while (dwSamples-- > 0)
    {
        INT16 sample = * (PINT16) pSrcBuffer;
        pDstBuffer[0] = sample;
        pDstBuffer[1] = sample;
        pDstBuffer += 2;
        pSrcBuffer += 1;

//DEBUGMSG(DBG_LVL_AC97,"%x\n", sample);
    }
}

STATIC VOID
BufferFillS16toS16 (PINT16 pDstBuffer, PINT16 pSrcBuffer, UINT32 dwSamples)
{
//DEBUGMSG(DBG_LVL_AC97,"BufferFillS16toS16\n");
    memcpy(pDstBuffer, pSrcBuffer, dwSamples * sizeof(UINT32));
}




UINT32
BufferFill( PWAVEHDR pWaveHdr, PVOID pvDstBuffer, UINT32 dwBufferSize )
{
    PBYTE   pDstBuffer = (PBYTE)pvDstBuffer;
    PBYTE   pOldDstBuffer = pDstBuffer;
    UINT32  dwDstSamples = dwBufferSize / ((gChannelDesc[WAPI_OUT].wDstBitsPerSample/8)*gChannelDesc[WAPI_OUT].wDstNbChannels);

//    DEBUGMSG(DBG_LVL_AC97,"+BufferFill\n");

    while ((pWaveHdr != NULL) && (dwDstSamples > 0))
    {
        if (pWaveHdr->nByteRecorded >= pWaveHdr->BufferLength)
        {
            if (pWaveHdr->nByteRecorded > pWaveHdr->BufferLength)
            {
                DEBUGMSG(DBG_LVL_AC97,"SHOULD NEVER HAPPEN !!! pWaveHdr nByteRecorded %d buf length %d\n",pWaveHdr->nByteRecorded,pWaveHdr->BufferLength);
            }
            pWaveHdr = pWaveHdr->pNext;
        }
        else {
            UINT32 dwSrcSamples = (pWaveHdr->BufferLength - pWaveHdr->nByteRecorded) / ((gChannelDesc[WAPI_OUT].wSrcBitsPerSample/8)*gChannelDesc[WAPI_OUT].wSrcNbChannels);
            UINT32 dwSamples = min(dwDstSamples, dwSrcSamples);
            gpfnBufferFill((PINT16) pDstBuffer, pWaveHdr->pBuffer + pWaveHdr->nByteRecorded, dwSamples);
            pWaveHdr->nByteRecorded += dwSamples * ((gChannelDesc[WAPI_OUT].wSrcBitsPerSample/8)*gChannelDesc[WAPI_OUT].wSrcNbChannels);
//            g_TotalPlayed += dwSamples;
            pDstBuffer += dwSamples * ((gChannelDesc[WAPI_OUT].wDstBitsPerSample/8)*gChannelDesc[WAPI_OUT].wDstNbChannels);
            dwDstSamples -= dwSamples;
        }
    }

    return (UINT32) pDstBuffer- (UINT32) pOldDstBuffer;
}




VOID
BufferGetM08fromM16 (PBYTE pDstBuffer, PINT16 pSrcBuffer, UINT32 dwSamples)
{
    while (dwSamples-- > 0)
    {
        *pDstBuffer = (BYTE) BITS_16_TO_8(pSrcBuffer[0]);
        pSrcBuffer += 1;
        pDstBuffer += 1;
    }
}

VOID
BufferGetS08fromM16 (PBYTE pDstBuffer, PINT16 pSrcBuffer, UINT32 dwSamples)
{
    register INT8 i8Value;
    while (dwSamples-- > 0)
    {
        i8Value = (BYTE) BITS_16_TO_8(pSrcBuffer[0]);
        pDstBuffer[0] = i8Value;
        pDstBuffer[1] = i8Value;
        pSrcBuffer += 1;
        pDstBuffer += 2;
    }
}


VOID
BufferGetM16fromM16 (PINT16 pDst, PINT16 pSrcBuffer, UINT32 dwSamples)
{
    memcpy(pDst, pSrcBuffer, dwSamples * sizeof(INT16));
}


VOID
BufferGetS16fromM16 (PINT16 pDstBuffer, PINT16 pSrcBuffer, UINT32 dwSamples)
{
    register INT16 i16Value;
    while (dwSamples-- > 0) 
    {
        i16Value = pSrcBuffer[0];
        pDstBuffer[0] = i16Value;
        pDstBuffer[1] = i16Value;
        pSrcBuffer += 1;
        pDstBuffer += 2;
    }
}




UINT32
BufferGet( PWAVEHDR pWaveHdr, PVOID pvSrcBuffer, UINT32 dwBufferSize )
{
    PBYTE pSrcBuffer = pvSrcBuffer;
    PVOID pOldSrcBuffer = pSrcBuffer;
    UINT32 dwSrcSamples = dwBufferSize / ((gChannelDesc[WAPI_IN].wSrcBitsPerSample/8)*gChannelDesc[WAPI_IN].wSrcNbChannels);
//DEBUGMSG(DBG_LVL_AC97,"+BufferGet\n");
    
    while ((pWaveHdr != NULL) && (dwSrcSamples > 0))
    {
        if (pWaveHdr->nByteRecorded >= pWaveHdr->BufferLength)
        {
            pWaveHdr = pWaveHdr->pNext;
        }
        else
        {
            UINT32 dwDstSamples = (pWaveHdr->BufferLength - pWaveHdr->nByteRecorded) / ((gChannelDesc[WAPI_IN].wDstBitsPerSample/8)*gChannelDesc[WAPI_IN].wDstNbChannels);
            UINT32 dwSamples = min(dwSrcSamples, dwDstSamples);
            gpfnBufferGet(pWaveHdr->pBuffer + pWaveHdr->nByteRecorded, (PINT16)pSrcBuffer, dwSamples);
            pWaveHdr->nByteRecorded += dwSamples * ((gChannelDesc[WAPI_IN].wDstBitsPerSample/8)*gChannelDesc[WAPI_IN].wDstNbChannels);
            pSrcBuffer += dwSamples * ((gChannelDesc[WAPI_IN].wSrcBitsPerSample/8)*gChannelDesc[WAPI_IN].wSrcNbChannels);
            dwSrcSamples -= dwSamples;
        }
    }
    
    return (UINT32) pSrcBuffer - (UINT32) pOldSrcBuffer;
}







//==============================================================================
// interface functions
//==============================================================================

PUBLIC BOOL
WavHal_Init( UINT32 Minor )
{
    int i;

    DEBUGMSG(DBG_LVL_AC97,"++WHAL_Init %X\n", Minor);

    //Initialize structure
    gChannelDesc[WAPI_OUT].bInUse = FALSE;
    gChannelDesc[WAPI_IN].bInUse = FALSE;

    if( !AC97_Initialize() )
        return FALSE;

    //Set the default values first
    gCaptureLatency = DEFAULT_CAPTURE_LATENCY;




    //Setup the buffers 
    for( i=0; i<NB_XMIT_PAGE; i++ )
    {
#if 0
        XmitBuffers[i].pData = DMAPage[i] -OFFSET_NONCACHE;
        XmitBuffers[i].dwPhysAddr = ((UINT32)DMAPage[i]) - OFFSET_NONCACHE;
#else
        XmitBuffers[i].pData = DMAPage[i];
        XmitBuffers[i].dwPhysAddr = ((UINT32)DMAPage[i]);
#endif
        XmitBuffers[i].ID = i;
    }

    for( i=0; i<NB_RECV_PAGE; i++ )
    {
#if 0
        RcvBuffers[i].pData = DMAPage[i + NB_XMIT_PAGE] - OFFSET_NONCACHE;
        RcvBuffers[i].dwPhysAddr = ((UINT32)DMAPage[i+ NB_XMIT_PAGE]) - OFFSET_NONCACHE;
#else
        RcvBuffers[i].pData = DMAPage[i + NB_XMIT_PAGE];
        RcvBuffers[i].dwPhysAddr = ((UINT32)DMAPage[i+ NB_XMIT_PAGE]);
#endif
        RcvBuffers[i].ID = i;
    }

    // Create the fifos
    BufferFifoInit(&FreeOutBufferFifo);
    BufferFifoInit(&BufferToPlayFifo);
    BufferFifoInit(&BufferBeingPlayedFifo);
    
    BufferFifoInit(&FreeInBufferFifo);
    BufferFifoInit(&BufferRecordedFifo);
    BufferFifoInit(&BufferBeingRecordedFifo);
        
    

    InterruptEnable(SYSINTR_WAV);

    DEBUGMSG(DBG_LVL_AC97,"--WHAL_Init\n");

    return TRUE;
}




PUBLIC AUDIO_STATE
WavHal_GetInterruptType( VOID )
{
    UINT32 dwCASR;
    AUDIO_STATE RetStatus;
    
    RetStatus=AUDIO_STATE_IGNORE ;
    dwCASR= g_pAC97C->AC97C_CASR & (g_pAC97C->AC97C_CAMR & (AT91C_AC97C_ENDRX | AT91C_AC97C_RXBUFF | AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_OVRUN | AT91C_AC97C_UNRUN));

//DEBUGMSG(DBG_LVL_AC97,"%s\n", __FUNCTION__);

    // for input (MIC-IN)
    if( gbMoreData[WAPI_IN] == TRUE )
    {
        if( dwCASR & (AT91C_AC97C_ENDRX | AT91C_AC97C_RXBUFF) )
        {
            T_BUFFER* pBuffer;
            UINT32 dwBothBufferConsummed;
            if (g_pAC97C->AC97C_CASR & AT91C_AC97C_RXBUFF)
            {
                pBuffer = BufferFifoPop(&BufferBeingRecordedFifo);
                if (pBuffer)
                {                                        
                    BufferFifoPush(&BufferRecordedFifo,pBuffer);
                }
                dwBothBufferConsummed = TRUE;
//DEBUGMSG(DBG_LVL_AC97,"ac97 RXBUFF\n");
            }
            else
            {
                dwBothBufferConsummed = FALSE;
//DEBUGMSG(DBG_LVL_AC97,"ac97 ENDRX\n");
            }
            pBuffer = BufferFifoPop(&BufferBeingRecordedFifo);
            if (pBuffer)
            {
                BufferFifoPush(&BufferRecordedFifo,pBuffer);
            }
            
            if (dwBothBufferConsummed)
            {
                T_BUFFER *pBuffer2;
                pBuffer = BufferFifoPop(&FreeInBufferFifo);                
                pBuffer2 = BufferFifoPop(&FreeInBufferFifo);
                
                if (pBuffer)
                {
                    g_pAC97C->AC97C_RPR = pBuffer->dwPhysAddr;
                    g_pAC97C->AC97C_RCR = pBuffer->dwSize /(gChannelDesc[WAPI_IN].wSrcBitsPerSample/8);
                    
                    BufferFifoPush(&BufferBeingRecordedFifo,pBuffer);
                    
                    if (pBuffer2)
                    {
//DEBUGMSG(DBG_LVL_AC97,"recording %d puis %d\n",pBuffer->ID,pBuffer2->ID);
                        g_pAC97C->AC97C_RNPR = pBuffer2->dwPhysAddr;
                        g_pAC97C->AC97C_RNCR = pBuffer2->dwSize /(gChannelDesc[WAPI_IN].wSrcBitsPerSample/8);
                        BufferFifoPush(&BufferBeingRecordedFifo,pBuffer2);
                    }
                    else
                    {
//DEBUGMSG(DBG_LVL_AC97,"recording %d\n",pBuffer->ID);
                    }
                    
                    //Enable the Rx interrupt
                    g_pAC97C->AC97C_CAMR |= AT91C_AC97C_ENDRX | AT91C_AC97C_RXBUFF | AT91C_AC97C_OVRUN;                                        
                    
                    RetStatus |= AUDIO_STATE_IN_RECORDING;
                }    
                else
                {
//DEBUGMSG(DBG_LVL_AC97,"int handler. No buffer to record in the Fifo\n");
                    //disable PDC TX interrupt 
                    g_pAC97C->AC97C_CAMR &= ~(AT91C_AC97C_ENDRX | AT91C_AC97C_RXBUFF | AT91C_AC97C_OVRUN);
                    RetStatus |= AUDIO_STATE_IN_RECORDING;          //TODO: Does the MDD now send a WPDM_STOP message?
                }
            }
            else
            {
                pBuffer = BufferFifoPop(&FreeInBufferFifo);

                if (pBuffer)
                {
//DEBUGMSG(DBG_LVL_AC97,"recording seulement %d \n",pBuffer->ID);
                    g_pAC97C->AC97C_RNPR = pBuffer->dwPhysAddr;
                    g_pAC97C->AC97C_RNCR = pBuffer->dwSize / (gChannelDesc[WAPI_IN].wSrcBitsPerSample/8);
                    
                    //Enable the Tx interrupt
                    g_pAC97C->AC97C_CAMR |= AT91C_AC97C_ENDRX | AT91C_AC97C_RXBUFF | AT91C_AC97C_OVRUN;                                        
                    
                    BufferFifoPush(&BufferBeingRecordedFifo,pBuffer);
                    
                    RetStatus |= AUDIO_STATE_IN_RECORDING;
                }    
                else
                {
//DEBUGMSG(DBG_LVL_AC97,"int handler. No buffer to record in the Fifo\n");
                    //disable PDC TX interrupt 
                    g_pAC97C->AC97C_CAMR &= ~(AT91C_AC97C_ENDRX | AT91C_AC97C_RXBUFF | AT91C_AC97C_OVRUN);                
                    RetStatus |= AUDIO_STATE_IN_RECORDING;          //TODO: Does the MDD now send a WPDM_STOP message?
                }
            }
            
        }
        if (dwCASR  & AT91C_AC97C_OVRUN)
        {
//DEBUGMSG(DBG_LVL_AC97,"PDD_AudioGetInterruptType : Receive Overrun\n");
            
            RetStatus |= AUDIO_STATE_IN_OVERFLOW;    // \bug the descriptor was loaded but not ready for use
        }
    }
    else
    {
        //disable PDC RX interrupt 
        g_pAC97C->AC97C_CAMR &= ~(AT91C_AC97C_ENDRX | AT91C_AC97C_RXBUFF | AT91C_AC97C_OVRUN);
        
        RetStatus |= AUDIO_STATE_IN_STOPPED;           //JJH TODO: Does the MDD now send a WPDM_STOP message?
    }




    // for output, (SOUNT-OUT )
    if( gbMoreData[WAPI_OUT] == TRUE )
    {
        if (dwCASR & (AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN)) //if the out channel stopped at the end of a buffer xfer
        {
            //BOOL dwBothBufferConsummed;
            T_BUFFER *  pBuffer;
            pBuffer = BufferFifoPop(&BufferBeingPlayedFifo);
            if (pBuffer)
            {
                gnXmitComplete += pBuffer->dwSize;
                pBuffer->dwSize = gChannelDesc[WAPI_OUT].dwDMASize;
                BufferFifoPush(&FreeOutBufferFifo,pBuffer);
//DEBUGMSG(DBG_LVL_AC97,"ID %d : %d \r\n",pBuffer->ID, pBuffer->dwSize);
            }
            
            if (g_pAC97C->AC97C_CASR & AT91C_AC97C_TXBUFE)
            {
                pBuffer = BufferFifoPop(&BufferBeingPlayedFifo);
                if (pBuffer)
                {
                    gnXmitComplete += pBuffer->dwSize;
                    pBuffer->dwSize = gChannelDesc[WAPI_OUT].dwDMASize;
                    BufferFifoPush(&FreeOutBufferFifo,pBuffer);
                }
                //dwBothBufferConsummed = TRUE;
//DEBUGMSG(DBG_LVL_AC97,"ac97 TXBUFE\n");
            }
            else
            if( g_pAC97C->AC97C_CASR & AT91C_AC97C_UNRUN )
            {
                // Underrun is manage as a TXBUFF.
                // When a under IT fired, driver consider all data in DMA as sent.
                g_pAC97C->AC97C_PTCR = AT91C_PDC_TXTDIS;
                pBuffer = BufferFifoPop(&BufferBeingPlayedFifo);
                if (pBuffer)
                {
                    gnXmitComplete += pBuffer->dwSize;
                    pBuffer->dwSize = gChannelDesc[WAPI_OUT].dwDMASize;
                    BufferFifoPush(&FreeOutBufferFifo,pBuffer);
                }
                //dwBothBufferConsummed = TRUE;
                g_pAC97C->AC97C_PTCR = AT91C_PDC_TXTEN;
            }
            else
            {
                //dwBothBufferConsummed = FALSE;
//DEBUGMSG(DBG_LVL_AC97,"ac97 ENDTX\n");
            }

#if 1
            RetStatus |= AUDIO_STATE_OUT_PLAYING;
#else
            if (dwBothBufferConsummed)
            {
                T_BUFFER *pBuffer2;
                pBuffer = BufferFifoPop(&BufferToPlayFifo);                
                pBuffer2 = BufferFifoPop(&BufferToPlayFifo);
                
                if (pBuffer)
                {
                    g_pAC97C->AC97C_TPR = pBuffer->dwPhysAddr;
                    g_pAC97C->AC97C_TCR = pBuffer->dwSize /(gChannelDesc[WAPI_OUT].wDstBitsPerSample/8);
                    
                    BufferFifoPush(&BufferBeingPlayedFifo,pBuffer);
                    
                    if (pBuffer2)
                    {
//DEBUGMSG(DBG_LVL_AC97,"playing %d plus %d\r\n",pBuffer->ID,pBuffer2->ID);
                        g_pAC97C->AC97C_TNPR = pBuffer2->dwPhysAddr;
                        g_pAC97C->AC97C_TNCR = pBuffer2->dwSize /(gChannelDesc[WAPI_OUT].wDstBitsPerSample/8);
                        BufferFifoPush(&BufferBeingPlayedFifo,pBuffer2);
                    }
                    else
                    {
//DEBUGMSG(DBG_LVL_AC97,"playing %d no secondary buffer\n",pBuffer->ID);
                    }
                    //Enable the Tx interrupt
                    g_pAC97C->AC97C_CAMR |= (AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);                
                    
                    RetStatus |= AUDIO_STATE_OUT_PLAYING;
                }    
                else
                {
//DEBUGMSG(DBG_LVL_AC97,"int handler. D No buffer to play in the Fifo\n");
                    //disable PDC TX interrupt 
                    g_pAC97C->AC97C_CAMR &= ~(AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);                
                    RetStatus |= AUDIO_STATE_OUT_PLAYING;          //TODO: Does the MDD now send a WPDM_STOP message?
                }
                
            }
            else
            {
                pBuffer = BufferFifoPop(&BufferToPlayFifo);

                if (pBuffer)
                {
//DEBUGMSG(DBG_LVL_AC97,"playing %d\n",pBuffer->ID);
                    g_pAC97C->AC97C_TNPR = pBuffer->dwPhysAddr;
                    g_pAC97C->AC97C_TNCR = pBuffer->dwSize / (gChannelDesc[WAPI_OUT].wDstBitsPerSample/8);
                    
                    //Enable the Tx interrupt
                    g_pAC97C->AC97C_CAMR |= (AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);
                    
                    BufferFifoPush(&BufferBeingPlayedFifo,pBuffer);
                    
                    RetStatus |= AUDIO_STATE_OUT_PLAYING;
                }    
                else
                {
//DEBUGMSG(DBG_LVL_AC97,"int handler. S No buffer to play in the Fifo\n");
                    //disable PDC TX interrupt 
                    g_pAC97C->AC97C_CAMR &= ~(AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);                
                    RetStatus |= AUDIO_STATE_OUT_PLAYING;          //TODO: Does the MDD now send a WPDM_STOP message?
                }
            }        

#endif

        }
        if (dwCASR  & AT91C_AC97C_UNRUN)
        {
//DEBUGMSG(DBG_LVL_AC97,"PDD_AudioGetInterruptType : Transmit Underrun\n");
            //RetStatus |= AUDIO_STATE_OUT_UNDERFLOW;    // Underrun is manage as a TXBUFF. See code below.
        }
    }
    else
    {
        //disable PDC TX interrupt 
        g_pAC97C->AC97C_CAMR &= ~(AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);
        
        RetStatus |= AUDIO_STATE_OUT_STOPPED;          //TODO: Does the MDD now send a WPDM_STOP message?
    }

    return RetStatus;
}




PUBLIC BOOL
WavHal_WaveOpen( WAPI_INOUT apidir, PWAVEFORMAT pFormat )
{

DEBUGMSG(DBG_LVL_AC97," pFormat = %X\n", pFormat);

DEBUGMSG(DBG_LVL_AC97," FormatTag = %d\n", pFormat->FormatTag);
DEBUGMSG(DBG_LVL_AC97," nChannel = %d\n", pFormat->nChannel);
DEBUGMSG(DBG_LVL_AC97," nSamplesPerSec = %d\n", pFormat->nSamplesPerSec);
DEBUGMSG(DBG_LVL_AC97," BitsPerSample = %d\n", pFormat->BitsPerSample);

    //invalid direction : no input audio source
    if( apidir >= NUM_API_DIRS )
    {
        DEBUGMSG(DBG_LVL_AC97,"Invalid Direction\r\n ");
        return FALSE;
    }

    if( pFormat->FormatTag != WAVE_FORMAT_PCM )
    {
        DEBUGMSG(DBG_LVL_AC97, "Invalid bad format\n");
        return FALSE;
    }
        
    //invalid bit per sample
    if( pFormat->BitsPerSample != 8 && pFormat->BitsPerSample != 16)
    {
        DEBUGMSG(DBG_LVL_AC97,"Invalid bits per sample\r\n ");
        return FALSE;
    }

    //invalid bit nbChannels
    if( pFormat->nChannel !=1  && pFormat->nChannel != 2)
    {
        DEBUGMSG(DBG_LVL_AC97,"Invalid number of channels\r\n ");
        return FALSE;
    }

    //invalid bit rate 
    if( pFormat->nSamplesPerSec != KHZ08_000 &&
        pFormat->nSamplesPerSec != KHZ11_025 &&
        pFormat->nSamplesPerSec != KHZ16_000 &&
        pFormat->nSamplesPerSec != KHZ22_050 &&
        pFormat->nSamplesPerSec != KHZ32_000 &&
        pFormat->nSamplesPerSec != KHZ44_100 &&
        pFormat->nSamplesPerSec != KHZ48_000 )
    {
        DEBUGMSG(DBG_LVL_AC97,"Invalid samples per second\r\n ");
        return FALSE;
    }

    //channel already in use
    if( gChannelDesc[apidir].bInUse )
    {
        DEBUGMSG(DBG_LVL_AC97, "channel already in use\r\n");
        return FALSE;
    }
    
    gChannelDesc[apidir].bInUse = TRUE;

    if (apidir == WAPI_OUT)
    {
        gChannelDesc[apidir].wSrcBitsPerSample = pFormat->BitsPerSample;
        gChannelDesc[apidir].wDstBitsPerSample = 16;
        gChannelDesc[apidir].wSrcNbChannels = pFormat->nChannel;
        gChannelDesc[apidir].wDstNbChannels = 2;
    }
    else
    {
        gChannelDesc[apidir].wSrcBitsPerSample = pFormat->BitsPerSample;
        gChannelDesc[apidir].wDstBitsPerSample = 16;
        gChannelDesc[apidir].wSrcNbChannels = pFormat->nChannel;
        gChannelDesc[apidir].wDstNbChannels = 1;
    }

    if( apidir == WAPI_OUT )
    {
        if (pFormat->BitsPerSample == 8)
        {
            if (pFormat->nChannel == 1)
                gpfnBufferFill = BufferFillM08toS16;
            else
                gpfnBufferFill = BufferFillS08toS16;
        }
        else
        {
            if (pFormat->nChannel == 1)
                gpfnBufferFill = BufferFillM16toS16;
            else
                gpfnBufferFill = BufferFillS16toS16;
        }
        gChannelDesc[apidir].dwDMASize = AUDIO_BUFFER_SIZE;
    }
    else
    {
        if( pFormat->BitsPerSample == 8 )
        {
            if (pFormat->nChannel == 1)
                gpfnBufferGet = BufferGetM08fromM16;
            else
                gpfnBufferGet = BufferGetS08fromM16;
        }
        else
        {
            if (pFormat->nChannel == 1)
                gpfnBufferGet = BufferGetM16fromM16;
            else
                gpfnBufferGet = BufferGetS16fromM16;
        }

        // Support Low-latency Capture
        // Adjust the effective DMA buffer size, based on the requested sample rate.
        // The idea is that we want our DMA capture buffers to represent a fixed amount of time,
        // regardless of the sample rate.
        // This means that if the application buffers are small (as is the case with VoIP or DirectSoundCapture)
        // then we still can return the buffer very soon after we have enough captured data to fill it.
        // The way this used to work (fixed-size capture buffers) meant that at low sample rates
        // we would wait for hundreds of milliseconds, then turn around and discover that the application
        // had only given us 80 milliseconds of wave headers, so we would be forced to drop samples on the floor.

        gChannelDesc[WAPI_IN].dwDMASize = ((UINT32) gChannelDesc[apidir].wSrcBitsPerSample / 8 * (UINT32) gChannelDesc[apidir].wSrcNbChannels * (UINT32)pFormat->nSamplesPerSec * (UINT32) gCaptureLatency /1000);
        if (gChannelDesc[WAPI_IN].dwDMASize > AUDIO_BUFFER_SIZE) 
            gChannelDesc[WAPI_IN].dwDMASize = AUDIO_BUFFER_SIZE;
    }


    //ENABLE VRA
    if( !ShadowWriteAC97( EXTENDED_AUDIO_CTRL, VRA_ENABLED_MASK)  )  // Enable Variable Rate Audio
    {
DEBUGMSG(DBG_LVL_AC97,"-- Bad VRA Value \n");
    }
    
    //set the sample rate
    if (AC97SetSampleRate((UINT16)pFormat->nSamplesPerSec,apidir) == FALSE)
    {
DEBUGMSG(DBG_LVL_AC97,"return bad sample rate:\n");
        return FALSE;
    }   


    // reset controller
    //ResetAC97Controller(CODEC_WARM_RESET);


    //Enable DAC or ADC
    if(apidir == WAPI_OUT)
    {
        bPowerDown[WAPI_OUT] = FALSE;
        AudioPowerDownUnit(PWR_PR1_DAC,FALSE);
        WaveOutMute(FALSE);
    }
    else
    {
        bPowerDown[WAPI_IN] = FALSE;
        WaveInMute(FALSE);   
        AudioPowerDownUnit(PWR_PR0_ADC,FALSE);
        Sleep(100);     //Give the Codec some time to settle
    }

    return TRUE;
    

}


PUBLIC VOID
WavHal_WaveOutStart( PWAVEHDR pWaveHdr )
{


    T_BUFFER* pBuffer;
    T_BUFFER* pBuffer2;
    UINT32 i;
    
DEBUGMSG(1,"++%s\n", __FUNCTION__);
//    g_TotalPlayed=0;


    // Reinit our fifos 
    BufferFifoReset(&FreeOutBufferFifo);
    BufferFifoReset(&BufferToPlayFifo);
    BufferFifoReset(&BufferBeingPlayedFifo);    
    for (i=0;i<NB_XMIT_PAGE;i++)
    {
        XmitBuffers[i].dwSize = gChannelDesc[WAPI_OUT].dwDMASize;
        BufferFifoPush(&FreeOutBufferFifo,&XmitBuffers[i]);
    }

    gbMoreData[WAPI_OUT] = TRUE;        // we expect to have more data coming...
    
#if 1
    pBuffer = BufferFifoPop(&FreeOutBufferFifo);
    if (pBuffer)
    {
        UINT32 dwSize = BufferFill(pWaveHdr, (PUINT32) pBuffer->pData, pBuffer->dwSize);
        pBuffer->dwSize = dwSize ; 
        BufferFifoPush(&BufferToPlayFifo,pBuffer);
    }
    pBuffer = BufferFifoPop(&FreeOutBufferFifo);
    if (pBuffer)
    {
        UINT32 dwSize = BufferFill(pWaveHdr, (PUINT32) pBuffer->pData, pBuffer->dwSize);
        pBuffer->dwSize = dwSize;
        BufferFifoPush(&BufferToPlayFifo,pBuffer);
    }
#else
    pBuffer = BufferFifoPop(&FreeOutBufferFifo);
    if (pBuffer)
    {
        MemSet( pBuffer->pData, 0 , gChannelDesc[WAPI_OUT].dwDMASize );
        pBuffer->dwSize = gChannelDesc[WAPI_OUT].dwDMASize ; 
        BufferFifoPush(&BufferToPlayFifo,pBuffer);
    }
    pBuffer = BufferFifoPop(&FreeOutBufferFifo);
    if (pBuffer)
    {
        MemSet( pBuffer->pData, 0 , gChannelDesc[WAPI_OUT].dwDMASize );
        pBuffer->dwSize = gChannelDesc[WAPI_OUT].dwDMASize ; 
        BufferFifoPush(&BufferToPlayFifo,pBuffer);
    }
#endif    
    
    pBuffer = BufferFifoPop(&BufferToPlayFifo);
    pBuffer2 = BufferFifoPop(&BufferToPlayFifo);
    
    if (pBuffer && pBuffer2)
    {
        g_pAC97C->AC97C_PTCR = AT91C_PDC_TXTDIS;
        g_pAC97C->AC97C_TPR = pBuffer->dwPhysAddr;                
        g_pAC97C->AC97C_TNPR = pBuffer2->dwPhysAddr;
        g_pAC97C->AC97C_TCR = pBuffer->dwSize / (gChannelDesc[WAPI_OUT].wDstBitsPerSample/8);            
        g_pAC97C->AC97C_TNCR = pBuffer2->dwSize / (gChannelDesc[WAPI_OUT].wDstBitsPerSample/8);            
        g_pAC97C->AC97C_PTCR = AT91C_PDC_TXTEN;
        BufferFifoPush(&BufferBeingPlayedFifo,pBuffer);
        BufferFifoPush(&BufferBeingPlayedFifo,pBuffer2);
    }

    if( bPowerDown[WAPI_OUT] == TRUE )
    {
        AudioPowerDownUnit(PWR_PR1_DAC,FALSE);
        bPowerDown[WAPI_OUT] = FALSE;
    }

    //Enable the PDC tx interrupt
    g_pAC97C->AC97C_CAMR |= (AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);


//DEBUGMSG(DBG_LVL_AC97,"--WavHal_WaveOutStart\n");
    return;
}



PUBLIC VOID
WavHal_WaveOutEndOfData( VOID )
{
    DEBUGMSG(1,"+WavHal_WaveOutEndOfData\n");

    //
    // There is no more data coming...
    //
    gbMoreData[WAPI_OUT] = FALSE;

    StopDmac(WAPI_OUT);

    return;
}

PUBLIC VOID
WavHal_WaveOutStop( VOID )
{
    DEBUGMSG(DBG_LVL_AC97,"+WavHal_WaveOutStop");

    //WavHal_WaveOutEndOfData();
    gbMoreData[WAPI_OUT] = FALSE;

    StopDmac(WAPI_OUT);

    return;
}


PUBLIC VOID
WavHal_WaveOutStandby( VOID )
{
    DEBUGMSG(DBG_LVL_AC97,"+WavHal_WaveOutStandby\n");

    // We are going to be idle. so power off what we can.

    bPowerDown[WAPI_OUT] = TRUE;
    AudioPowerDownUnit(PWR_PR1_DAC,TRUE);  //turn
    //AudioPowerDownUnit(PWR_PR0_ADC,PowerDownIn);  //turn
        
}

PUBLIC BOOL
WavHal_WaveOutPause ( VOID )
{
    DEBUGMSG(DBG_LVL_AC97,"+WavHal_WaveOutPause\n");

    WavHal_WaveOutEndOfData();

    return TRUE;
}

PUBLIC BOOL
WavHal_WaveOutRestart( PWAVEHDR pWaveHdr )
{
    DEBUGMSG(DBG_LVL_AC97,"+WavHal_WaveOutRestart\n");

    WavHal_WaveOutStart (pWaveHdr);

    return TRUE;
}



PUBLIC UINT32
WavHal_WaveOutContinue( PWAVEHDR pWaveHdr )
{
    T_BUFFER * pBuffer;
    BOOL bLoop = TRUE;
    
//DEBUGMSG(DBG_LVL_AC97,"+WavHal_WaveOutContinue\n");

//    private_AudioComplete(pWaveHdr, (gnXmitComplete * gChannelDesc.wSrcBitsPerSample * gChannelDesc.wSrcNbChannels) / (gChannelDesc.wDstBitsPerSample* gChannelDesc.wDstNbChannels));

    gnXmitComplete = 0;
    
    do
    {        
        pBuffer = BufferFifoPop(&FreeOutBufferFifo);
        if (pBuffer)
        {
            UINT32 dwSize;                        
            
            dwSize = BufferFill(pWaveHdr, (PUINT32) pBuffer->pData, pBuffer->dwSize);
DEBUGMSG(DBG_LVL_AC97,"ID %d : %d / %d\r\n",pBuffer->ID, pBuffer->dwSize,dwSize);
            if (pBuffer->dwSize != dwSize) //Is the buffer entirely filled ?            
            {                
//DEBUGMSG(DBG_LVL_AC97,"ID %d : %d / %d\r\n",pBuffer->ID, pBuffer->dwSize,dwSize);
                //The buffer hasn't filled entirely, then we have no more data available
                bLoop = FALSE;
            }
            
                
            if (dwSize)
            {
                pBuffer->dwSize = dwSize;
                BufferFifoPush(&BufferToPlayFifo,pBuffer);
            }
#if 1
            else
            {

                MemSet( pBuffer->pData, 0 , gChannelDesc[WAPI_OUT].dwDMASize );
                pBuffer->dwSize = gChannelDesc[WAPI_OUT].dwDMASize ; 
                BufferFifoPush(&BufferToPlayFifo,pBuffer);
            }
#else
            else
            {
                BufferFifoPush(&FreeOutBufferFifo,pBuffer);
            }
#endif
        }
    } while (bLoop && pBuffer);



#if 1
    {
        T_BUFFER *pBuffer2;

        pBuffer = BufferFifoPop(&BufferToPlayFifo);                
        pBuffer2 = BufferFifoPop(&BufferToPlayFifo);

        if( pBuffer && pBuffer2 )
        {
            g_pAC97C->AC97C_TPR = pBuffer->dwPhysAddr;
            g_pAC97C->AC97C_TCR = pBuffer->dwSize /(gChannelDesc[WAPI_OUT].wDstBitsPerSample/8);
            BufferFifoPush(&BufferBeingPlayedFifo,pBuffer);
            
            g_pAC97C->AC97C_TNPR = pBuffer2->dwPhysAddr;
            g_pAC97C->AC97C_TNCR = pBuffer2->dwSize /(gChannelDesc[WAPI_OUT].wDstBitsPerSample/8);
            BufferFifoPush(&BufferBeingPlayedFifo,pBuffer2);
        }
        else
        {
            if (pBuffer)
            {
//DEBUGMSG(DBG_LVL_AC97,"playing %d\n",pBuffer->ID);
                g_pAC97C->AC97C_TNPR = pBuffer->dwPhysAddr;
                g_pAC97C->AC97C_TNCR = pBuffer->dwSize / (gChannelDesc[WAPI_OUT].wDstBitsPerSample/8);
                
                //Enable the Tx interrupt
                g_pAC97C->AC97C_CAMR |= (AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);
                
                BufferFifoPush(&BufferBeingPlayedFifo,pBuffer);
            }    
            else
            {
//DEBUGMSG(DBG_LVL_AC97,"int handler. S No buffer to play in the Fifo\n");
                //disable PDC TX interrupt 
                g_pAC97C->AC97C_CAMR &= ~(AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);                
            }
        }        
    }

#else
            if (dwBothBufferConsummed)
            {
                T_BUFFER *pBuffer2;
                pBuffer = BufferFifoPop(&BufferToPlayFifo);                
                pBuffer2 = BufferFifoPop(&BufferToPlayFifo);
                
                if (pBuffer)
                {
                    g_pAC97C->AC97C_TPR = pBuffer->dwPhysAddr;
                    g_pAC97C->AC97C_TCR = pBuffer->dwSize /(gChannelDesc[WAPI_OUT].wDstBitsPerSample/8);
                    
                    BufferFifoPush(&BufferBeingPlayedFifo,pBuffer);
                    
                    if (pBuffer2)
                    {
//DEBUGMSG(DBG_LVL_AC97,"playing %d plus %d\r\n",pBuffer->ID,pBuffer2->ID);
                        g_pAC97C->AC97C_TNPR = pBuffer2->dwPhysAddr;
                        g_pAC97C->AC97C_TNCR = pBuffer2->dwSize /(gChannelDesc[WAPI_OUT].wDstBitsPerSample/8);
                        BufferFifoPush(&BufferBeingPlayedFifo,pBuffer2);
                    }
                    else
                    {
DEBUGMSG(DBG_LVL_AC97,"playing %d no secondary buffer\n",pBuffer->ID);
                    }
                    //Enable the Tx interrupt
                    g_pAC97C->AC97C_CAMR |= (AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);                
                    
                    RetStatus |= AUDIO_STATE_OUT_PLAYING;
                }    
                else
                {
DEBUGMSG(DBG_LVL_AC97,"int handler. D No buffer to play in the Fifo\n");
                    //disable PDC TX interrupt 
                    g_pAC97C->AC97C_CAMR &= ~(AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);                
                    RetStatus |= AUDIO_STATE_OUT_PLAYING;          //TODO: Does the MDD now send a WPDM_STOP message?
                }
                
            }
            else
            {
                pBuffer = BufferFifoPop(&BufferToPlayFifo);

                if (pBuffer)
                {
//DEBUGMSG(DBG_LVL_AC97,"playing %d\n",pBuffer->ID);
                    g_pAC97C->AC97C_TNPR = pBuffer->dwPhysAddr;
                    g_pAC97C->AC97C_TNCR = pBuffer->dwSize / (gChannelDesc[WAPI_OUT].wDstBitsPerSample/8);
                    
                    //Enable the Tx interrupt
                    g_pAC97C->AC97C_CAMR |= (AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);
                    
                    BufferFifoPush(&BufferBeingPlayedFifo,pBuffer);
                    
                    RetStatus |= AUDIO_STATE_OUT_PLAYING;
                }    
                else
                {
DEBUGMSG(DBG_LVL_AC97,"int handler. S No buffer to play in the Fifo\n");
                    //disable PDC TX interrupt 
                    g_pAC97C->AC97C_CAMR &= ~(AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);                
                    RetStatus |= AUDIO_STATE_OUT_PLAYING;          //TODO: Does the MDD now send a WPDM_STOP message?
                }
            }        

#endif




    //Enable the PDC tx interrupt .. just in case it had been disabled (no more data available)
    g_pAC97C->AC97C_CAMR |= (AT91C_AC97C_ENDTX | AT91C_AC97C_TXBUFE | AT91C_AC97C_UNRUN);


    return 0;
}


PUBLIC UINT32
WavHal_WaveOutClose( VOID )
{
    DEBUGMSG(DBG_LVL_AC97,"+WavHal_WaveOutClose\n");

    AC97_Deinitialize();

    return 0;    
}





//------------------------------------------------------------------------------------------------------------
//! \brief the WPDM_START message where the WAPI_INOUT direction is WAPI_IN.
//-------------------------------------------------------------------------------------------------------------
PUBLIC VOID
WavHal_WaveInStart( PWAVEHDR pWaveHdr )
{        
    T_BUFFER* pBuffer;
    T_BUFFER* pBuffer2;
    UINT32 i;

    DEBUGMSG(DBG_LVL_AC97,"+WavHal_WaveInStart\n");

    BufferFifoReset(&FreeInBufferFifo);
    BufferFifoReset(&BufferRecordedFifo);
    BufferFifoReset(&BufferBeingRecordedFifo);

    for (i=0;i<NB_XMIT_PAGE;i++)
    {
        RcvBuffers[i].dwSize = gChannelDesc[WAPI_IN].dwDMASize;
        BufferFifoPush(&FreeInBufferFifo,&RcvBuffers[i]);
    }

    //clear the mute
    WaveInMute(FALSE);     

    gbMoreData[WAPI_IN]  = TRUE;        // we expect to have more data coming...

    pBuffer = BufferFifoPop(&FreeInBufferFifo);
    pBuffer2 = BufferFifoPop(&FreeInBufferFifo);
    
    if (pBuffer && pBuffer2)
    {
        g_pAC97C->AC97C_PTCR = AT91C_PDC_RXTDIS;
        g_pAC97C->AC97C_RPR = pBuffer->dwPhysAddr;                
        g_pAC97C->AC97C_RNPR = pBuffer2->dwPhysAddr;
        g_pAC97C->AC97C_RCR = pBuffer->dwSize / (gChannelDesc[WAPI_IN].wDstBitsPerSample/8);            
        g_pAC97C->AC97C_RNCR = pBuffer2->dwSize / (gChannelDesc[WAPI_IN].wDstBitsPerSample/8);            
        g_pAC97C->AC97C_PTCR = AT91C_PDC_RXTEN;        
        BufferFifoPush(&BufferBeingRecordedFifo,pBuffer);
        BufferFifoPush(&BufferBeingRecordedFifo,pBuffer2);
    }        

    //Enable the Rx interrupt
    g_pAC97C->AC97C_CAMR |= AT91C_AC97C_ENDRX | AT91C_AC97C_RXBUFF | AT91C_AC97C_OVRUN;

    return;    
}

//------------------------------------------------------------------------------------------------------------
//! \brief  instruct the audio PDD to continue with the data pointed to by the wave header
//-------------------------------------------------------------------------------------------------------------
PUBLIC VOID
WavHal_WaveInContinue( PWAVEHDR pWaveHdr )
{    
    T_BUFFER * pBuffer;
    UINT32 dwSize;
    BOOL bLoop = TRUE;
    
    DEBUGMSG(DBG_LVL_AC97, "+WavHal_WaveInContinue\n");
    
    do
    {
        pBuffer = BufferFifoPop(&BufferRecordedFifo);
        if (pBuffer)
        {            
//DEBUGMSG(DBG_LVL_AC97,"getting buffer %d\n",pBuffer->ID);
            dwSize = BufferGet(pWaveHdr, pBuffer->pData, pBuffer->dwSize);
            
            BufferFifoPush(&FreeInBufferFifo,pBuffer);            
        }
        bLoop = FALSE;
    } while (pBuffer && bLoop);
    
    //Enable the Rx interrupt
    g_pAC97C->AC97C_CAMR |= AT91C_AC97C_ENDRX | AT91C_AC97C_RXBUFF | AT91C_AC97C_OVRUN;
    
    return;
}


PUBLIC VOID
WavHal_WaveInStandby( VOID )
{
    DEBUGMSG(DBG_LVL_AC97,"+WavHal_WaveInStandby\n");

    // We are going to be idle. so power off what we can.

    bPowerDown[WAPI_IN] = TRUE;
    AudioPowerDownUnit(PWR_PR0_ADC, TRUE);  //turn
        
}

//------------------------------------------------------------------------------------------------------------
//! \brief stop recording immediately
//-------------------------------------------------------------------------------------------------------------
PUBLIC VOID
WavHal_WaveInStop(PWAVEHDR pWaveHdr )
{
    T_BUFFER * pBuffer;

    DEBUGMSG(DBG_LVL_AC97,"+WavHal_WaveInStop\n");
    
    gbMoreData[WAPI_IN] = FALSE;
    
    StopDmac(WAPI_IN);
    
    WaveInMute(TRUE);      
    
    do
    {
        pBuffer = BufferFifoPop(&BufferRecordedFifo);
        if (pBuffer)
        {
            BufferGet(pWaveHdr, pBuffer->pData, pBuffer->dwSize);
            BufferFifoPush(&FreeInBufferFifo,pBuffer);
        }
    } while (pBuffer);

    return;    
}





PUBLIC BOOL 
WavHal_AudioMessage(UINT32 Msg, UINT32 Param1, UINT32 Param2)
{
    DEBUGMSG(DBG_LVL_AC97,"+WavHal_AudioMessage\n");
    switch( Msg )
    {
        case WODM_GET_VOLUME:
            *((PUINT32) Param1) = v_nVolume;
            break;
        
        case WODM_SET_VOLUME:
            v_nVolume = (UINT32) Param1;
            WaveOutSetVolume(v_nVolume);		
            break;

        default:
            break;
    }

    return TRUE;
}


